@@ -8,7 +8,7 @@
  *             Andrae Muys <andrae@humbug.org.au.au>
  *           - uses libbzip2 by Julian Seward (http://sources.redhat.com/bzip2/)
  *
- *	Date  : March 20, 2007
+ *	Date  : July 25, 2007
  *
  *
  *  Contributions
@@ -30,19 +30,24 @@
  *  Anbal Monsalve Salazar <anibal@debian.org> - creates and maintains Debian packages
  *  Steve Christensen - creates and maintains Solaris packages (sunfreeware.com)
  *  Alessio Cervellin - creates and maintains Solaris packages (blastwave.org)
+ *  Ying-Chieh Liao - created the FreeBSD port
+ *  Andrew Pantyukhin <sat@FreeBSD.org> - maintains the FreeBSD ports and willing to
+ *          resolve any FreeBSD-related problems
+ *  Roland Illig <rillig@NetBSD.org> - creates and maintains NetBSD packages
  *
  *  Specials thanks for suggestions and testing:  Phillippe Welsh,
  *  James Terhune, Dru Lemley, Bryan Stillwell, George Chalissery,
  *  Kir Kolyshkin, Madhu Kangara, Mike Furr, Joergen Ramskov, Kurt Fitzner,
  *  Peter Cordes, Oliver Falk, Jindrich Novy, Benjamin Reed, Chris Dearman,
- *  Richard Russon, Anbal Monsalve Salazar, Jim Leonard, Paul Pluzhnikov.
+ *  Richard Russon, Anbal Monsalve Salazar, Jim Leonard, Paul Pluzhnikov,
+ *  Coran Fisher, Ken Takusagawa, David Pyke.
  *
  *
  * This program, "pbzip2" is copyright (C) 2003-2007 Jeff Gilchrist.
  * All rights reserved.
  *
  * The library "libbzip2" which pbzip2 uses, is copyright
- * (C) 1996-2002 Julian R Seward.  All rights reserved.
+ * (C) 1996-2006 Julian R Seward.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -77,7 +82,7 @@
  *
  * Jeff Gilchrist, Ottawa, Canada.
  * pbzip2@compression.ca
- * pbzip2 version 1.0.1 of March 20, 2007
+ * pbzip2 version 1.0.2 of July 25, 2007
  *
  */
 #include <vector>
@@ -128,6 +133,7 @@
 #endif
 
 #ifdef WIN32
+#define PATH_SEP	'\\'
 #define usleep(x) Sleep(x/1000)
 #define LOW_DWORD(x)  ((*(LARGE_INTEGER *)&x).LowPart)
 #define HIGH_DWORD(x) ((*(LARGE_INTEGER *)&x).HighPart)
@@ -138,6 +144,8 @@ struct timeval {
   long tv_usec;
 };
 #endif
+#else
+#define PATH_SEP	'/'
 #endif
 
 #ifndef WIN32
@@ -203,6 +211,7 @@ char *memstr(char *, int, char *, int);
 int producer_decompress(int, OFF_T, queue *);
 void *consumer_decompress(void *);
 void *fileWriter(void *);
+int directcompress(int, OFF_T, int, char *);
 int producer(int, int, OFF_T, int, queue *);
 void *consumer(void *);
 queue *queueInit(int);
@@ -833,6 +842,186 @@ void *fileWriter(void *outname)
 	return (NULL);
 }
 
+/*
+ *********************************************************
+ */
+int directcompress(int hInfile, OFF_T fileSize, int blockSize, char *OutFilename)
+{
+	char *FileData = NULL;
+	char *CompressedData = NULL;
+	OFF_T CompressedSize = 0;
+	OFF_T bytesLeft = 0;
+	OFF_T inSize = 0;
+	unsigned int outSize = 0;
+	int percentComplete = 0;
+	int hOutfile = 1;  // default to stdout
+	int currBlock = 0;
+	int rret = 0;
+	int ret = 0;
+
+	bytesLeft = fileSize;
+
+	// write to file instead of stdout
+	if (OutputStdOut == 0)
+	{
+		hOutfile = open(OutFilename, O_RDWR | O_CREAT | O_TRUNC | O_BINARY, FILE_MODE);
+		// check to see if file creation was successful
+		if (hOutfile == -1)
+		{
+			fprintf(stderr, " *ERROR: Could not create output file [%s]!\n", OutFilename);
+			return -1;
+		}
+	}
+
+	// keep going until all the file is processed
+	while (bytesLeft > 0)
+	{
+		//
+		// READ DATA
+		//
+		
+		// set buffer size
+		if (bytesLeft > blockSize)
+			inSize = blockSize;
+		else
+			inSize = bytesLeft;
+
+		#ifdef PBZIP_DEBUG
+		fprintf(stderr, " -> Bytes To Read: %"PRIu64" bytes...\n", inSize);
+		#endif
+
+		// allocate memory to read in file
+		FileData = NULL;
+		FileData = new char[inSize];
+		// make sure memory was allocated properly
+		if (FileData == NULL)
+		{
+			fprintf(stderr, " *ERROR: Could not allocate memory (FileData)!  Skipping...\n");
+			close(hInfile);
+			if (OutputStdOut == 0)
+				close(hOutfile);
+			allDone = 1;
+			return -1;
+		}
+
+		// read file data
+		rret = read(hInfile, (char *) FileData, inSize);
+		#ifdef PBZIP_DEBUG
+		fprintf(stderr, " -> Total Bytes Read: %d bytes...\n\n", rret);
+		#endif
+		if (rret == 0)
+		{
+			if (FileData != NULL)
+				delete [] FileData;
+			break;
+		}
+		else if (rret < 0)
+		{
+			fprintf(stderr, " *ERROR: Could not read from file!  Skipping...\n");
+			close(hInfile);
+			if (FileData != NULL)
+				delete [] FileData;
+			if (OutputStdOut == 0)
+				close(hOutfile);
+			allDone = 1;
+			return -1;
+		}
+
+		// set bytes left after read
+		bytesLeft -= rret;
+
+		//
+		// COMPRESS DATA
+		//
+			
+		outSize = (int) ((inSize*1.01)+600);
+		// allocate memory for compressed data
+		CompressedData = new char[outSize];
+		// make sure memory was allocated properly
+		if (CompressedData == NULL)
+		{
+			fprintf(stderr, " *ERROR: Could not allocate memory (CompressedData)!  Skipping...\n");
+			close(hInfile);
+			if (FileData != NULL)
+				delete [] FileData;
+			allDone = 1;
+			return -1;
+		}
+
+		// compress the memory buffer (blocksize=9*100k, verbose=0, worklevel=30)
+		ret = BZ2_bzBuffToBuffCompress(CompressedData, &outSize, FileData, inSize, BWTblockSize, Verbosity, 30);
+		if (ret != BZ_OK)
+			fprintf(stderr, " *ERROR during compression: %d\n", ret);
+
+		#ifdef PBZIP_DEBUG
+		fprintf(stderr, "\n   Original Block Size: %u\n", inSize);
+		fprintf(stderr, " Compressed Block Size: %u\n", outSize);
+		#endif
+
+		//
+		// WRITE DATA
+		//
+
+		// write data to the output file
+		ret = write(hOutfile, CompressedData, outSize);
+
+		#ifdef PBZIP_DEBUG
+		fprintf(stderr, "\n -> Total Bytes Written[%d]: %d bytes...\n", currBlock, ret);
+		#endif
+		CompressedSize += ret;
+		if (ret <= 0)
+		{
+			fprintf(stderr, " *ERROR: Could not write to file!  Skipping...\n");
+			close(hInfile);
+			if (FileData != NULL)
+				delete [] FileData;
+			if (CompressedData != NULL)
+				delete [] CompressedData;
+			if (OutputStdOut == 0)
+				close(hOutfile);
+			allDone = 1;
+			return -1;
+		}
+
+		currBlock++;
+		// print current completion status
+		percentComplete = 100 * currBlock / NumBlocks;
+		if (QuietMode != 1)
+		{
+			fprintf(stderr, "Completed: %d%%             \r", percentComplete);
+			fflush(stderr);
+		}
+		
+		// clean up memory
+		if (FileData != NULL)
+		{
+			delete [] FileData;
+			FileData = NULL;
+		}
+		if (CompressedData != NULL)
+		{
+			delete [] CompressedData;
+			CompressedData = NULL;
+		}
+		
+		// check to make sure all the data we expected was read in
+		if (rret != inSize)
+			inSize = rret;
+	} // while
+
+	close(hInfile);
+	
+	if (OutputStdOut == 0)
+		close(hOutfile);
+	if (QuietMode != 1)
+	{
+		fprintf(stderr, "    Output Size: %"PRIu64" bytes\n", (unsigned long long)CompressedSize);
+	}
+
+	allDone = 1;
+	return 0;
+}
+
 /*
  *********************************************************
  */
@@ -1405,8 +1594,8 @@ int writeFileMetaData(char *fileName)
  */
 void banner()
 {
-	fprintf(stderr, "Parallel BZIP2 v1.0.1 - by: Jeff Gilchrist [http://compression.ca]\n");
-	fprintf(stderr, "[Mar. 20, 2007]             (uses libbzip2 by Julian Seward)\n");
+	fprintf(stderr, "Parallel BZIP2 v1.0.2 - by: Jeff Gilchrist [http://compression.ca]\n");
+	fprintf(stderr, "[July 25, 2007]             (uses libbzip2 by Julian Seward)\n");
 
 	return;
 }
@@ -1453,6 +1642,8 @@ int main(int argc, char* argv[])
 	pthread_t output;
 	char **FileList = NULL;
 	char *InFilename = NULL;
+	char *progName = NULL;
+	char *progNamePos = NULL;
 	char bz2Header[] = {"BZh91AY&SY"};  // using 900k block size
 	char OutFilename[255];
 	char cmdLineTemp[255];
@@ -1486,12 +1677,13 @@ int main(int argc, char* argv[])
 	int blockSize = 9*100000;
 	int decompress = 0;
 	int testFile = 0;
+	int errLevel = 0;
+	int noThreads = 0;
 	int keep = 0;
 	int force = 0;
 	int ret = 0;
-	int i, j, k;
 	int fileLoop;
-	int errLevel = 0;
+	int i, j, k;
 	
 	// get current time for benchmark reference
 	#ifndef WIN32
@@ -1509,6 +1701,16 @@ int main(int argc, char* argv[])
 		usage(argv[0]);
 	}
 
+	// get program name to determine if decompress mode should be used
+	progName = argv[0];
+	for (progNamePos = argv[0]; progNamePos[0] != '\0'; progNamePos++)
+	{
+		if (progNamePos[0] == PATH_SEP)
+			progName = progNamePos + 1;
+	}
+	if ((strstr(progName, "unzip") != 0) || (strstr(progName, "UNZIP") != 0))
+		decompress = 1;
+	
 	FileListCount = 0;
 	FileList = new char *[argc];
 	if (FileList == NULL)
@@ -1782,6 +1984,7 @@ int main(int argc, char* argv[])
 			{
 				fprintf(stderr, " *ERROR: File [%s] NOT found!  Skipping...\n", InFilename);
 				fprintf(stderr, "-------------------------------------------\n");
+				errLevel = 1;
 				continue;
 			}
 			memset(tmpBuff, 0, sizeof(tmpBuff));
@@ -1791,6 +1994,7 @@ int main(int argc, char* argv[])
 			{
 				fprintf(stderr, " *ERROR: File [%s] is NOT a valid bzip2!  Skipping...\n", InFilename);
 				fprintf(stderr, "-------------------------------------------\n");
+				errLevel = 1;
 				continue;
 			}
 			else
@@ -1800,6 +2004,7 @@ int main(int argc, char* argv[])
 				{
 					fprintf(stderr, " *ERROR: File [%s] is NOT a valid bzip2!  Skipping...\n", InFilename);
 					fprintf(stderr, "-------------------------------------------\n");
+					errLevel = 1;
 					continue;
 				}
 				// skip 4th char which differs depending on BWT block size used
@@ -1807,6 +2012,7 @@ int main(int argc, char* argv[])
 				{
 					fprintf(stderr, " *ERROR: File [%s] is NOT a valid bzip2!  Skipping...\n", InFilename);
 					fprintf(stderr, "-------------------------------------------\n");
+					errLevel = 1;
 					continue;
 				}
 				// set block size for decompression
@@ -1816,6 +2022,7 @@ int main(int argc, char* argv[])
 				{
 					fprintf(stderr, " *ERROR: File [%s] is NOT a valid bzip2!  Skipping...\n", InFilename);
 					fprintf(stderr, "-------------------------------------------\n");
+					errLevel = 1;
 					continue;
 				}
 			}
@@ -1839,6 +2046,7 @@ int main(int argc, char* argv[])
 			{
 				fprintf(stderr, " *ERROR: Input file [%s] already has a .bz2 extension!  Skipping...\n", InFilename);
 				fprintf(stderr, "-------------------------------------------\n");
+				errLevel = 1;
 				continue;
 			}
 			strcat(OutFilename, ".bz2");
@@ -1855,6 +2063,7 @@ int main(int argc, char* argv[])
 		{
 			fprintf(stderr, " *ERROR: File [%s] NOT found!  Skipping...\n", InFilename);
 			fprintf(stderr, "-------------------------------------------\n");
+			errLevel = 1;
 			continue;
 		}
 
@@ -1865,6 +2074,7 @@ int main(int argc, char* argv[])
 		{
 			fprintf(stderr, " *ERROR: File [%s] is a directory!  Skipping...\n", InFilename);
 			fprintf(stderr, "-------------------------------------------\n");
+			errLevel = 1;
 			continue;
 		}
 		// check to make sure input is a regular file
@@ -1872,6 +2082,7 @@ int main(int argc, char* argv[])
 		{
 			fprintf(stderr, " *ERROR: File [%s] is not a regular file!  Skipping...\n", InFilename);
 			fprintf(stderr, "-------------------------------------------\n");
+			errLevel = 1;
 			continue;
 		}
 		// get size of file
@@ -1886,6 +2097,7 @@ int main(int argc, char* argv[])
 		{
 			fprintf(stderr, " *ERROR: File is of size 0 [%s]!  Skipping...\n", InFilename);
 			fprintf(stderr, "-------------------------------------------\n");
+			errLevel = 1;
 			continue;
 		}
 
@@ -1894,6 +2106,7 @@ int main(int argc, char* argv[])
 		{
 			fprintf(stderr, " *ERROR: Could not get file meta data from [%s]!  Skipping...\n", InFilename);
 			fprintf(stderr, "-------------------------------------------\n");
+			errLevel = 1;
 			continue;
 		}
 
@@ -1906,7 +2119,9 @@ int main(int argc, char* argv[])
 			{
 				fprintf(stderr, " *ERROR: Output file [%s] already exists!  Use -f to overwrite...\n", OutFilename);
 				fprintf(stderr, "-------------------------------------------\n");
+				errLevel = 1;
 				close(hOutfile);
+				errLevel = 1;
 				continue;
 			}
 		}
@@ -1935,17 +2150,25 @@ int main(int argc, char* argv[])
 		if (decompress == 1)
 		{
 			numBlocks = 0;
+			noThreads = 0;
 		}
 		else
 		{
 			// calculate the # of blocks of data
 			numBlocks = (fileSize + blockSize - 1) / blockSize;
+			// Do not use threads for small files where we only have 1 block to process
+			// or if we only have 1 CPU
+			if ((numBlocks == 1) || (numCPU == 1))
+				noThreads = 1;
+			else
+				noThreads = 0;
 		}
 		#ifdef PBZIP_DEBUG
 		fprintf(stderr, "# Blocks: %d\n", numBlocks);
 		#endif
 		// set global variable
 		NumBlocks = numBlocks;
+		
 		// create output buffer
 		OutputBuffer.resize(numBlocks);
 		// make sure memory was allocated properly
@@ -1963,6 +2186,7 @@ int main(int argc, char* argv[])
 
 		if (decompress == 1)
 		{
+			// do decompression
 			NumBufferedBlocks = 0;
 			if (QuietMode != 1)
 				fprintf(stderr, "Decompressing data...\n");
@@ -1984,39 +2208,63 @@ int main(int argc, char* argv[])
 			}
 
 			// start reading in data for decompression
-			producer_decompress(hInfile, fileSize, fifo);
+			ret = producer_decompress(hInfile, fileSize, fifo);
+			if (ret != 0)
+				errLevel = 1;
 		}
 		else
 		{
-			if (QuietMode != 1)
-				fprintf(stderr, "Compressing data...\n");
-			for (i=0; i < numCPU; i++)
+			// do compression code
+				
+			// use multi-threaded code
+			if (noThreads == 0)
 			{
-				ret = pthread_create(&fifo->consumers[i], NULL, consumer, fifo);
+				if (QuietMode != 1)
+					fprintf(stderr, "Compressing data...\n");
+					
+				for (i=0; i < numCPU; i++)
+				{
+					ret = pthread_create(&fifo->consumers[i], NULL, consumer, fifo);
+					if (ret != 0)
+					{
+						fprintf(stderr, " *ERROR: Not enough resources to create consumer thread #%d (code = %d)  Aborting...\n", i, ret);
+						return 1;
+					}
+				}
+	
+				ret = pthread_create(&output, NULL, fileWriter, OutFilename);
 				if (ret != 0)
 				{
-					fprintf(stderr, " *ERROR: Not enough resources to create consumer thread #%d (code = %d)  Aborting...\n", i, ret);
+					fprintf(stderr, " *ERROR: Not enough resources to create fileWriter thread (code = %d)  Aborting...\n", ret);
 					return 1;
 				}
+	
+				// start reading in data for compression
+				ret = producer(hInfile, fileSize, blockSize, fifo);
+				if (ret != 0)
+					errLevel = 1;
 			}
-
-			ret = pthread_create(&output, NULL, fileWriter, OutFilename);
-			if (ret != 0)
+			else
 			{
-				fprintf(stderr, " *ERROR: Not enough resources to create fileWriter thread (code = %d)  Aborting...\n", ret);
-				return 1;
+				// do not use threads for compression
+				if (QuietMode != 1)
+					fprintf(stderr, "Compressing data (no threads)...\n");
+
+				ret = directcompress(hInfile, fileSize, blockSize, OutFilename);
+				if (ret != 0)
+					errLevel = 1;
 			}
+		} // else
 
-			// start reading in data for compression
-			producer(hInfile, fileSize, blockSize, fifo);
-		}
-
-		// wait for fileWriter thread to exit
-		pthread_join(output, NULL);
-
-		// wait for consumer threads to exit
-		for (i = 0; i < numCPU; i++)
-		  pthread_join(fifo->consumers[i], NULL);
+		if (noThreads == 0)
+		{
+			// wait for fileWriter thread to exit
+			pthread_join(output, NULL);
+	
+			// wait for consumer threads to exit
+			for (i = 0; i < numCPU; i++)
+			  pthread_join(fifo->consumers[i], NULL);
+	  	}
 
 		if (OutputStdOut == 0)
 		{
